re_processedname = r'^_?[A-Z\._]*$'


class StoreFactory:

    def __del__(self):
        self.close()


    def add(self, k):
        raise NotImplementedError()


    def close(self):
        pass


    def ls(self):
        raise NotImplementedError()



# standard imports
import datetime

# external imports
import redis

# local imports
from .base import StoreFactory


class RedisStore:

    def __init__(self, path, redis, binary=False):
        self.redis = redis
        self.__path = path
        self.__binary = binary


    def __to_path(self, k):
        return '.'.join([self.__path, k])


    def __from_path(self, s):
        (left, right) = s.split(b'.', maxsplit=1)
        return right


    def __to_result(self, v):
        if self.__binary:
            return v
        return v.decode('utf-8')


    def put(self, k, contents=b''):
        if contents == None:
            contents = b''
        k = self.__to_path(k)
        self.redis.set(k, contents)


    def remove(self, k):
        k = self.__to_path(k)
        self.redis.delete(k)


    def get(self, k):
        k = self.__to_path(k)
        v = self.redis.get(k)
        return self.__to_result(v)

    
    def list(self):
        (cursor, matches) = self.redis.scan(match=self.__path + '.*')

        r = []
        for s in matches:
            k = self.__from_path(s)
            v = self.redis.get(k)
            r.append((k, v,))

        return r


    def path(self):
        return None


    def replace(self, k, contents):
        if contents == None:
            contents = b''
        k = self.__to_path(k)
        v = self.redis.get(k)
        if v == None:
            raise FileNotFoundError(k)
        self.redis.set(k, contents)


    def modified(self, k):
        k = self.__to_path(k)
        k = '_mod' + k
        v = self.redis.get(k)
        return int(v)


    def register_modify(self, k):
        k = self.__to_path(k)
        k = '_mod' + k
        ts = datetime.datetime.utcnow().timestamp()
        self.redis.set(k)


class RedisStoreFactory(StoreFactory):

    def __init__(self, host='localhost', port=6379, db=2, binary=False):
        self.redis = redis.Redis(host=host, port=port, db=db)
        self.__binary = binary


    def add(self, k):
        k = str(k)
        return RedisStore(k, self.redis, binary=self.__binary)


    def close(self):
        self.redis.close()


    def ls(self):
        r = []
        (c, ks) = self.redis.scan(match='*')
        for k in ks:
            v = k.rsplit(b'.', maxsplit=1)
            if v != k:
                v = v[0].decode('utf-8')
                if v not in r:
                    r.append(v)
        return r



# standard imports
import datetime
import os

# external imports
import rocksdb

# local imports
from .base import StoreFactory


class RocksDbStore:

    def __init__(self, path, db, binary=False):
        self.db = db
        self.__path = path
        self.__binary = binary


    def __to_key(self, k):
        return k.encode('utf-8')


    def __to_contents(self, v):
        if isinstance(v, bytes):
            return v
        return v.encode('utf-8')


    def __to_path(self, k):
        return '.'.join([self.__path, k])


    def __from_path(self, s):
        (left, right) = s.split('.', maxsplit=1)
        return right


    def __to_result(self, v):
        if self.__binary:
            return v
        return v.decode('utf-8')


    def put(self, k, contents=b''):
        if contents == None:
            contents = b''
        else:
            contents = self.__to_contents(contents)
        k = self.__to_path(k)
        k = self.__to_key(k)
        self.db.put(k, contents)


    def remove(self, k):
        k = self.__to_path(k)
        k = self.__to_key(k)
        self.db.delete(k)


    def get(self, k):
        k = self.__to_path(k)
        k = self.__to_key(k)
        v = self.db.get(k)
        return self.__to_result(v)

 
    def list(self):
        it = self.db.iteritems()
        kb_start = self.__to_key(self.__path)
        it.seek(kb_start)

        r = []
        l = len(self.__path)
        for (kb, v) in it:
            k = kb.decode('utf-8') 
            if len(k) < l or k[:l] != self.__path:
                break
            k = self.__from_path(k)
            v = self.db.get(kb)
            r.append((k, v,))

        return r


    def path(self):
        return None


    def replace(self, k, contents):
        if contents == None:
            contents = b''
        else:
            contents = self.__to_contents(contents)
        k = self.__to_path(k)
        k = self.__to_key(k)
        v = self.db.get(k)
        if v == None:
            raise FileNotFoundError(k)
        self.db.put(k, contents)


    def modified(self, k):
        k = self.__to_path(k)
        k = '_mod' + k
        v = self.db.get(k)
        return int(v)


    def register_modify(self, k):
        k = self.__to_path(k)
        k = '_mod' + k
        ts = datetime.datetime.utcnow().timestamp()
        self.db.set(k)


class RocksDbStoreFactory(StoreFactory):

    def __init__(self, path, binary=False):
        try:
            os.stat(path)
        except FileNotFoundError:
            os.makedirs(path)
        self.db = rocksdb.DB(path, rocksdb.Options(create_if_missing=True))
        self.__binary = binary


    def add(self, k):
        k = str(k)
        return RocksDbStore(k, self.db, binary=self.__binary)


    def close(self):
        self.db.close()


    def ls(self):
        it = self.db.iterkeys()
        it.seek_to_first()
        r = []
        for k in it:
            v = k.rsplit(b'.', maxsplit=1)
            if v != k:
                v = v[0].decode('utf-8')
                if v not in r:
                    r.append(v)
        return r



# standard imports
import datetime

# local imports
from .state import State
from .error import (
        StateItemExists,
        StateLockedKey,
        )


class PersistedState(State):
    """Adapter for persisting state changes and synchronising states between memory and persisted backend.

    :param factory: A function capable of returning a persisted store from a single path argument.
    :type factory: function
    :param bits: Number of pure states. Passed to the superclass.
    :type bits: int
    :param logger: Logger to capture logging output, or None for no logging.
    :type logger: object
    """

    def __init__(self, factory, bits, logger=None, verifier=None, check_alias=True, event_callback=None):
        super(PersistedState, self).__init__(bits, logger=logger, verifier=verifier, check_alias=check_alias, event_callback=event_callback)
        self.__store_factory = factory
        self.__stores = {}


    # Create state store container if missing.
    def __ensure_store(self, k):
        if self.__stores.get(k) == None:
            self.__stores[k] = self.__store_factory(k)


    def put(self, key, contents=None, state=None):
        """Persist a key or key/content pair.

        See shep.state.State.put
        """
        k = self.to_name(state)

        self.__ensure_store(k)

        self.__stores[k].put(key, contents)

        super(PersistedState, self).put(key, state=state, contents=contents)

        self.register_modify(key)


    def set(self, key, or_state):
        """Persist a new state for a key or key/content.

        See shep.state.State.set
        """
        from_state = self.state(key)
        k_from = self.name(from_state)

        to_state = super(PersistedState, self).set(key, or_state)
        k_to = self.name(to_state)
        self.__ensure_store(k_to)

        contents = None
        try:
            contents = self.__stores[k_from].get(key)
            self.__stores[k_to].put(key, contents)
            self.__stores[k_from].remove(key)
        except StateLockedKey as e:
            super(PersistedState, self).unset(key, or_state, allow_base=True)
            raise e
       
        #self.sync(to_state)

        return to_state


    def unset(self, key, not_state):
        """Persist a new state for a key or key/content.

        See shep.state.State.unset
        """
        from_state = self.state(key)
        k_from = self.name(from_state)

        to_state = super(PersistedState, self).unset(key, not_state)

        k_to = self.name(to_state)
        self.__ensure_store(k_to)

        contents = self.__stores[k_from].get(key)
        self.__stores[k_to].put(key, contents)
        self.__stores[k_from].remove(key)

        return to_state


    def change(self, key, bits_set, bits_unset):
        """Persist a new state for a key or key/content.

        See shep.state.State.unset
        """
        from_state = self.state(key)
        k_from = self.name(from_state)

        to_state = super(PersistedState, self).change(key, bits_set, bits_unset)

        k_to = self.name(to_state)
        self.__ensure_store(k_to)

        contents = self.__stores[k_from].get(key)
        self.__stores[k_to].put(key, contents)
        self.__stores[k_from].remove(key)

        self.register_modify(key)

        return to_state


    def move(self, key, to_state):
        """Persist a new state for a key or key/content.

        See shep.state.State.move
        """
        from_state = self.state(key)
        to_state = super(PersistedState, self).move(key, to_state)
        return self.__movestore(key, from_state, to_state)


    # common procedure for safely moving a persisted resource from one state to another.
    def __movestore(self, key, from_state, to_state):
        k_from = self.name(from_state)
        k_to = self.name(to_state)

        self.__ensure_store(k_to)

        contents = self.__stores[k_from].get(key)
        self.__stores[k_to].put(key, contents)
        self.__stores[k_from].remove(key)

        self.register_modify(key)

        self.sync(to_state)

        return to_state


    def sync(self, state=None, not_state=None):
        """Reload resources for a single state in memory from the persisted state store.

        :param state: State to load
        :type state: int
        :raises StateItemExists: A content key is already recorded with a different state in memory than in persisted store.
        # :todo: if sync state is none, sync all
        """

        states_numeric = []
        if state == None:
            states_numeric = list(self.all(numeric=True))
        else:
            states_numeric = [state]
       
        states = []
        for state in states_numeric:
            if not_state != None:
                if state & not_state == 0:
                    states.append(self.name(state))
            else:
                states.append(self.name(state))

        ks = []
        for k in states:
            ks.append(k)

        for k in ks:
            self.__ensure_store(k)
            for o in self.__stores[k].list():
                state = self.from_name(k)
                try:
                    super(PersistedState, self).put(o[0], state=state, contents=o[1])
                except StateItemExists as e:
                    pass


    def list(self, state):
        """List all content keys for a particular state.

        This method will return from memory, and will not sync the persisted state first.
   
        See shep.state.State.list
        """
        k = self.name(state)
        self.__ensure_store(k)
        return super(PersistedState, self).list(state)


    def path(self, state, key=None):
        """Return a file path or URL pointing to the persisted state.
        
        If the key is omitted, the URL to the state item's container must be returned, and None if no such container exists.
         
        :param state: State to locate
        :type state: int
        :param key: Content key to locate
        :type key: str
        :rtype: str
        :returns: Locator pointng to persisted state
        :todo: rename to "location"
        """
        k = self.name(state)
        self.__ensure_store(k)
        return self.__stores[k].path(k=key)


    def next(self, key=None):
        """Advance and persist to the next pure state.

        See shep.state.State.next
        """
        from_state = self.state(key)
        to_state = super(PersistedState, self).next(key)
        return self.__movestore(key, from_state, to_state)


    def replace(self, key, contents):
        """Replace contents associated by content key.

        See shep.state.State.replace
        """
        state = self.state(key)
        k = self.name(state)
        r = self.__stores[k].replace(key, contents)
        super(PersistedState, self).replace(key, contents)
        return r


    def modified(self, key):
        state = self.state(key)
        k = self.name(state)
        return self.__stores[k].modified(key)



class StateExists(Exception):
    """Attempt to add state that already exists.
    """
    pass


class StateInvalid(Exception):
    """Attempt to operate on or move to a state that does not exist.
    """
    pass


class StateItemExists(Exception):
    """A content key attempted added that already exists.
    """
    pass


class StateItemNotFound(Exception):
    """A content key attempted read that does not exist.
    """
    pass


class StateCorruptionError(RuntimeError):
    """An irrecoverable discrepancy between persisted state and memory state has occurred.
    """
    pass


class StateTransitionInvalid(Exception):
    """Raised if state transition verification fails
    """
    pass


class StateLockedKey(Exception):
    """Attempt to write to a state key that is being written to by another client
    """
    pass



# standard imports
import re
import datetime

# local imports
from shep.error import (
        StateExists,
        StateInvalid,
        StateItemExists,
        StateItemNotFound,
        StateTransitionInvalid,
        StateCorruptionError,
        )


re_name = r'^[a-zA-Z_\.]+$'

class State:
    """State is an in-memory bitmasked state store for key-value pairs, or even just keys alone.

    A State is comprised of a number of atomic state bits, and zero or more aliases that represent unique combinations of these bits.

    The State object will enforce that duplicate states cannot exist. It will also enforce that all alias states are composed of valid atomic states.

    :param bits: Number of atomic states that this State object will represent (i.e. number of bits).
    :type bits: int
    :param logger: Standard library logging instance to output to
    :type logger: logging.Logger
    """

    base_state_name = 'NEW'

    def __init__(self, bits, logger=None, verifier=None, check_alias=True, event_callback=None):
        self.__initial_bits = bits
        self.__bits = bits
        self.__limit = (1 << bits) - 1
        self.__c = 0
        setattr(self, self.base_state_name, 0)

        self.__reverse = {0: getattr(self, self.base_state_name)}
        self.__keys = {getattr(self, self.base_state_name): []}
        self.__keys_reverse = {}
        self.__contents = {}
        self.modified_last = {}
        self.verifier = verifier
        self.check_alias = check_alias
        self.event_callback = event_callback


    @classmethod
    def set_default_state(cls, state_name):
        cls.base_state_name = state_name.upper()


    # return true if v is a single-bit state
    def __is_pure(self, v):
        if v == 0:
            return True
        c = 1
        for i in range(self.__bits):
            if c & v > 0:
                break
            c <<= 1
        return c == v


    # validates a state name and return its canonical representation
    def __check_name_valid(self, k):
        if not re.match(re_name, k):
            raise ValueError('only alpha and underscore')
        return k.upper()


    # enforces name validity, aswell as name uniqueness
    def __check_name(self, k):
        k = self.__check_name_valid(k) 
            
        try:
            getattr(self, k)
            raise StateExists(k)
        except AttributeError:
            pass
        return k


    # enforces state value validity and uniqueness
    def __check_valid(self, v):
        v = self.__check_value_typ(v)
        if self.__reverse.get(v):
            raise StateExists(v)
        return v


    # enforces state value within bit limit of instantiation
    def __check_limit(self, v, pure=True):
        if pure:
            if self.__initial_bits == 0:
                self.__bits += 1
            self.__limit = (1 << self.__bits) - 1
        if v > self.__limit:
            raise OverflowError(v)
        return v


    # enforces state value validity, uniqueness and value limit 
    def __check_value(self, v):
        v = self.__check_valid(v)
        self.__check_limit(v) 
        return v


    # enforces state value validity
    def __check_value_typ(self, v):
        return int(v)


    # enforces state value validity within the currently registered states (number of add calls vs number of bits in instantiation).
    def __check_value_cursor(self, v):
        v = self.__check_value_typ(v)
        if v > 1 << self.__c:
            raise StateInvalid(v)
        return v


    # set a bit for state of the given key
    def __set(self, k, v):
        setattr(self, k, v)
        self.__reverse[v] = k
        self.__c += 1


    # check validity of key to register state for
    def __check_key(self, item):
        if self.__keys_reverse.get(item) != None:
            raise StateItemExists(item)


    # adds a new key to the state store
    def __add_state_list(self, state, item):
        if self.__keys.get(state) == None:
            self.__keys[state] = []
        if not self.__is_pure(state) or state == 0:
            self.__keys[state].append(item)
        c = 1
        for i in range(self.__bits):
            part = c & state
            if part > 0:
                if self.__keys.get(part) == None:
                    self.__keys[part] = []
                self.__keys[part].append(item)
            c <<= 1
        self.__keys_reverse[item] = state
        if self.__reverse.get(state) == None and not self.check_alias:
            s = self.elements(state)
            self.__alias(s, state)


    def __state_list_index(self, item, state_list):
        """Get index of a key for a given state.
        A key should only ever exist in one state.
        A failed lookup should indicate a mistake on the caller part, (it may also indicate corruption, but probanbly impossible to tell the difference)
        """
        idx = -1
        try:
            idx = state_list.index(item)
        except ValueError:
            pass

        if idx == -1:
            raise StateCorruptionError() # should have state int here as value

        return idx


    def add(self, k):
        """Add a state to the store.
        
        :param k: State name
        :type k: str
        :raises shep.error.StateExists: State name is already registered
        """
        v = 1 << self.__c
        k = self.__check_name(k)
        v = self.__check_value(v)
        self.__set(k, v)


    def to_name(self, k):
        if k == None:
            k = 0
        return self.name(k)


    def __alias(self, k, *args):
        v = 0
        for a in args:
            a = self.__check_value_cursor(a)
            v = self.__check_limit(v | a, pure=False)
        if self.__is_pure(v):
            raise ValueError('use add to add pure values')
        return self.__set(k, v)


    def alias(self, k, *args):
        """Add an alias for a combination of states in the store.
        
        State aggregates may be provided as comma separated values or as a single (or'd) integer value. 
        
        :param k: Alias name
        :type k: str
        :param *args: One or more states to aggregate for this alias.
        :type *args: int or list of ints
        :raises StateInvalid: Attempt to create alias for one or more atomic states that do not exist.
        :raises ValueError: Attempt to use bit value as alias
        """
        k = self.__check_name(k)
        return self.__alias(k, *args)    


    def all(self, pure=False, numeric=False):
        """Return list of all unique atomic and alias state strings.
        
        :rtype: list of ints
        :return: states
        """
        l = []
        for k in dir(self):
            state = None
            if k[0] == '_':
                continue
            if k.upper() != k:
                continue
            if pure:
                state = self.from_name(k)
                if not self.__is_pure(state):
                    continue
            if numeric:
                if state == None:
                    state = self.from_name(k)
                l.append(state)
            else:
                l.append(k)
        l.sort()
        return l


    def elements(self, v):
        r = []
        if v == None or v == 0:
            return self.base_state_name
        c = 1
        for i in range(self.__bits):
            if v & c > 0:
                r.append(self.name(c))
            c <<= 1
        return '_' + '.'.join(r)


    def from_elements(self, k):
        r = 0
        if k[0] != '_':
            raise ValueError('elements string must start with underscore (_), got {}'.format(k))
        for v in k[1:].split('.'):
            r |= self.from_name(v) 
        return r


    def name(self, v):
        """Retrieve that string representation of the state attribute represented by the given state integer value.
        
        :param v: State integer
        :type v: int
        :raises StateInvalid: State corresponding to given integer not found
        :rtype: str
        :return: State name
        """
        k = self.__reverse.get(v)
        if k == None:
            if self.check_alias:
                raise StateInvalid(v)
            else:
                k = self.elements(v)
        elif v == None or v == 0:
            return self.base_state_name
        return k


    def from_name(self, k):
        """Retrieve the real state integer value corresponding to an attribute name.
        
        :param k: Attribute name
        :type k: str
        :raises ValueError: Invalid attribute name
        :raises AttributeError: Attribute not found
        :rtype: int
        :return: Numeric state value
        """
        k = self.__check_name_valid(k)
        return getattr(self, k)


    def match(self, v, pure=False):
        """Match against all stored states.
        
        If pure is set, only match against the single atomic state will be returned.
        
        :param v: Integer state to match
        :type v: int
        :param pure: Match only pure states
        :type pure: bool
        :raises KeyError: Unknown state
        :rtype: tuple
        :return: 0: Alias that input resolves to, 1: list of atomic states that matches the state
        """
        alias = None
        if not pure:
            alias = self.__reverse.get(v)

        r = []
        c = 1
        for i in range(self.__bits):
            if v & c > 0:
                try:
                    k = self.__reverse[c]
                    r.append(k)
                except KeyError:
                    pass
            c <<= 1

        return (alias, r,)

   
    def put(self, key, state=None, contents=None):
        """Add a key to an existing state.
        
        If no state it specified, the default state attribute State.base_state_name will be used.
        
        Contents may be supplied as value to pair with the given key. Contents may be changed later by calling the `replace` method.
        
        :param key: Content key to add
        :type key: str
        :param state: Initial state for the put. If not given, initial state will be State.base_state_name
        :type state: int
        :param contents: Contents to associate with key. A valie of None should be recognized as an undefined value as opposed to a zero-length value throughout any backend
        :type contents: str
        :raises StateItemExists: Content key has already been added
        :raises StateInvalid: Given state has not been registered
        :rtype: integer
        :return: Resulting state that key is put under (should match the input state)
        """
        if state == None:
            state = getattr(self, self.base_state_name)
        elif self.__reverse.get(state) == None and self.check_alias:
            raise StateInvalid(state)
        self.__check_key(key)

        if self.event_callback != None:
            old_state = self.__keys_reverse.get(key)
            self.event_callback(key, None, self.name(state))

        self.__add_state_list(state, key)
        if contents != None:
            self.__contents[key] = contents

        self.register_modify(key)

        return state
                                

    def move(self, key, to_state):
        """Move a given content key from one state to another.
        
        :param key: Key to move
        :type key: str
        :param to_state: Numeric state to move to (may be atomic or alias)
        :type to_state: integer
        :raises StateItemNotFound: Given key has not been registered
        :raises StateInvalid: Given state has not been registered
        :rtype: integer
        :return: Resulting state from move (should match the state given as input)
        """
        current_state = self.__keys_reverse.get(key)
        if current_state == None:
            raise StateItemNotFound(key)

        new_state = self.__reverse.get(to_state)
        if new_state == None and self.check_alias:
            raise StateInvalid(to_state)

        return self.__move(key, current_state, to_state)


    # implementation for state move that ensures integrity of keys and states.
    def __move(self, key, from_state, to_state):
        current_state_list = self.__keys.get(from_state)
        if current_state_list == None:
            raise StateCorruptionError(current_state)

        idx = self.__state_list_index(key, current_state_list)

        new_state_list = self.__keys.get(to_state)
        if current_state_list == None:
            raise StateCorruptionError(to_state)

        if self.verifier != None:
            r = self.verifier(self, from_state, to_state)
            if r != None:
                raise StateTransitionInvalid(r)

        current_state_list.pop(idx) 

        if self.event_callback != None:
            old_state = self.__keys_reverse.get(key)
            self.event_callback(key, self.name(old_state), self.name(to_state))

        self.__add_state_list(to_state, key)

        self.register_modify(key)

        return to_state
   

    def set(self, key, or_state):
        """Move to an alias state by setting a single bit.
        
        :param key: Content key to modify state for
        :type key: str
        :param or_state: Atomic stat to add
        :type or_state: int
        :raises ValueError: State is not a single bit state
        :raises StateItemNotFound: Content key is not registered
        :raises StateInvalid: Resulting state after addition of atomic state is unknown
        :rtype: int
        :returns: Resulting state
        """
        if not self.__is_pure(or_state):
            raise ValueError('can only apply using single bit states')

        current_state = self.__keys_reverse.get(key)
        if current_state == None:
            raise StateItemNotFound(key)

        to_state = current_state | or_state
        new_state = self.__reverse.get(to_state)
        if new_state == None and self.check_alias:
            raise StateInvalid('resulting to state is unknown: {}'.format(to_state))

        return self.__move(key, current_state, to_state)


    def unset(self, key, not_state, allow_base=False):
        """Unset a single bit, moving to a pure or alias state.
        
        The resulting state cannot be State.base_state_name (0).
         
        :param key: Content key to modify state for
        :type key: str
        :param or_state: Atomic stat to add
        :type or_state: int
        :raises ValueError: State is not a single bit state, or attempts to revert to State.base_state_name
        :raises StateItemNotFound: Content key is not registered
        :raises StateInvalid: Resulting state after addition of atomic state is unknown
        :rtype: int
        :returns: Resulting state
        """
        if not self.__is_pure(not_state):
            raise ValueError('can only apply using single bit states')

        current_state = self.__keys_reverse.get(key)
        if current_state == None:
            raise StateItemNotFound(key)

        to_state = current_state & (~not_state)
        if to_state == current_state:
            raise ValueError('invalid change for state {}: {}'.format(key, not_state))

        if to_state == getattr(self, self.base_state_name) and not allow_base:
            raise ValueError('State {} for {} cannot be reverted to {}'.format(current_state, key, self.base_state_name))

        new_state = self.__reverse.get(to_state)
        if new_state == None:
            raise StateInvalid('resulting to state is unknown: {}'.format(to_state))

        return self.__move(key, current_state, to_state)


    def change(self, key, sets, unsets):
        current_state = self.__keys_reverse.get(key)
        if current_state == None:
            raise StateItemNotFound(key)
        to_state = current_state | sets
        to_state &= ~unsets & self.__limit

        if sets == 0:
            to_state = current_state & (~unsets)
            if to_state == current_state:
                raise ValueError('invalid change by unsets for state {}: {}'.format(key, unsets))

        if to_state == getattr(self, self.base_state_name):
            raise ValueError('State {} for {} cannot be reverted to {}'.format(current_state, key, self.base_state_name))

        new_state = self.__reverse.get(to_state)
        if new_state == None:
            raise StateInvalid('resulting to state is unknown: {}'.format(to_state))

        return self.__move(key, current_state, to_state)


    def state(self, key):
        """Return the current numeric state for the given content key.
        
        :param key: Key to return content for
        :type key: str
        :raises StateItemNotFound: Content key is unknown
        :rtype: int
        :returns: State
        """
        state = self.__keys_reverse.get(key)
        if state == None:
            raise StateItemNotFound(key)
        return state


    def get(self, key):
        """Retrieve the content for a content key.
        
        :param key: Content key to retrieve content for
        :type key: str
        :rtype: any
        :returns: Content
        """
        return self.__contents.get(key)


    def list(self, state):
        """List all content keys matching a state.
        
        :param state: State to match 
        :type state: int
        :rtype: list of str
        :returns: Matching content keys
        """
        try:
            return self.__keys[state]
        except KeyError:
            return []


    def sync(self, state=None):
        """Noop method for interface implementation providing sync to backend.
        
        :param state: State to sync.
        :type state:
        :todo: (for higher level implementer) if sync state is none, sync all
        """
        pass


    def path(self, state, key=None):
        """In the memory-only class no persisted state is used, and this will return None.
        
        See shep.persist.PersistedState.path for more information.
        """
        return None


    def peek(self, key):
        """Return the next pure state.
         
        Will return the same result as the method next, but without advancing to the new state.
        
        :param key: Content key to inspect state for
        :type key: str
        :raises StateItemNotFound: Unknown content key
        :raises StateInvalid: Attempt to advance from an alias state, OR beyond the last known pure state.
        :rtype: int
        :returns: Next state
        """
        state = self.__keys_reverse.get(key)
        if state == None:
            raise StateItemNotFound(key)
        if not self.__is_pure(state):
            raise StateInvalid('cannot run next on an alias state')
       
        if state == 0:
            state = 1
        else:
            state <<= 1
        if state > self.__limit:
            raise StateInvalid('unknown state {}'.format(state))

        return state


    def next(self, key):
        """Advance to the next pure state.
           
        :param key: Content key to inspect state for
        :type key: str
        :raises StateItemNotFound: Unknown content key
        :raises StateInvalid: Attempt to advance from an alias state, OR beyond the last known pure state.
        :rtype: int
        :returns: Next state
        """
        from_state = self.state(key)
        new_state = self.peek(key)
        return self.__move(key, from_state, new_state)


    def replace(self, key, contents):
        """Replace contents associated by content key.
         
        :param key: Content key to replace for
        :type key: str
        :param contents: New contents
        :type contents: any
        :raises KeyError: Unknown content key
        """
        self.state(key)
        self.__contents[key] = contents


    def modified(self, key):
        return self.modified_last[key]


    def register_modify(self, key):
        self.modified_last[key] = datetime.datetime.utcnow().timestamp()


    def mask(self, key, states=0):
        statemask = self.__limit + 1
        statemask |= states
        statemask = ~statemask
        statemask &= self.__limit
        return statemask


    def purge(self, key):
        state = self.state(key)
        state_name = self.name(state)

        v = self.__keys.get(state)
        v.remove(key) 

        del self.__keys_reverse[key]

        try:
            del self.__contents[key]
        except KeyError:
            pass

        try:
            del self.modified_last[key]
        except KeyError:
            pass



# standard imports
import os
import re

# local imports
from .base import (
        re_processedname,
        StoreFactory,
        )
from shep.error import StateLockedKey


class SimpleFileStore:
    """Filesystem store of contents for state, with one directory per state.

    :param path: Filesystem base path for all state directory
    :type path: str
    """
    def __init__(self, path, binary=False, lock_path=None):
        self.__path = path
        os.makedirs(self.__path, exist_ok=True)
        if binary:
            self.__m = ['rb', 'wb']
        else:
            self.__m = ['r', 'w']
        self.__lock_path = lock_path
        if self.__lock_path != None:
            os.makedirs(lock_path, exist_ok=True)


    def __lock(self, k):
        if self.__lock_path == None:
            return
        fp = os.path.join(self.__lock_path, k)
        f = None
        try:
            f = open(fp, 'x')
        except FileExistsError:
            pass
        if f == None:
            raise StateLockedKey(k)
        f.close()


    def __unlock(self, k):
        if self.__lock_path == None:
            return
        fp = os.path.join(self.__lock_path, k)
        try:
            os.unlink(fp)
        except FileNotFoundError:
            pass
        

    def put(self, k, contents=None):
        """Add a new key and optional contents 

        :param k: Content key to add 
        :type k: str
        :param contents: Optional contents to assign for content key
        :type contents: any
        """
        self.__lock(k)
        fp = os.path.join(self.__path, k)
        if contents == None:
            if self.__m[1] == 'wb':
                contents = b''
            else:
                contents = ''

        f = open(fp, self.__m[1])
        f.write(contents)
        f.close()
        self.__unlock(k)


    def remove(self, k):
        """Remove a content key from a state.

        :param k: Content key to remove from the state
        :type k: str
        :raises FileNotFoundError: Content key does not exist in the state
        """
        self.__lock(k)
        fp = os.path.join(self.__path, k)
        os.unlink(fp)
        self.__unlock(k)

    
    def get(self, k):
        """Retrieve the content for the given content key.

        :param k: Content key to retrieve content for
        :type k: str
        :raises FileNotFoundError: Content key does not exist for the state
        :rtype: any
        :return: Contents
        """
        self.__lock(k)
        fp = os.path.join(self.__path, k)
        f = open(fp, self.__m[0])
        r = f.read()
        f.close()
        self.__unlock(k)
        return r


    def list(self):
        """List all content keys persisted for the state.

        :rtype: list of str
        :return: Content keys in state
        """
        self.__lock('.list')
        files = []
        for p in os.listdir(self.__path):
            fp = os.path.join(self.__path, p)
            f = None
            try:
                f = open(fp, self.__m[0])
            except FileNotFoundError:
                continue
            r = f.read()
            f.close()
            if len(r) == 0:
                r = None
            files.append((p, r,))
        self.__unlock('.list')
        return files


    def path(self, k=None):
        """Return filesystem path for persisted state or state item.

        :param k: If given, will return filesystem path to specified content key
        :type k: str
        :rtype: str
        :return: File path
        """
        if k == None:
            return self.__path
        return os.path.join(self.__path, k)


    def replace(self, k, contents):
        """Replace persisted content for persisted content key.

        :param k: Content key to replace contents for
        :type k: str
        :param contents: Contents
        :type contents: any
        """
        self.__lock(k)
        fp = os.path.join(self.__path, k)
        os.stat(fp)
        f = open(fp, self.__m[1])
        r = f.write(contents)
        f.close()
        self.__unlock(k)


    def modified(self, k):
        self.__lock(k)
        path = self.path(k)
        st = os.stat(path)
        self.__unlock(k)
        return st.st_ctime


    def register_modify(self, k):
        pass


class SimpleFileStoreFactory(StoreFactory):
    """Provide a method to instantiate SimpleFileStore instances that provide persistence for individual states.

    :param path: Filesystem path as base path for states
    :type path: str
    """
    def __init__(self, path, binary=False, use_lock=False):
        self.__path = path
        self.__binary = binary
        self.__use_lock = use_lock


    def add(self, k):
        """Create a new SimpleFileStore for a state.

        :param k: Identifier for the state
        :type k: str
        :rtype: SimpleFileStore
        :return: A filesystem persistence instance with the given identifier as subdirectory
        """
        lock_path = None
        if self.__use_lock:
            lock_path = os.path.join(self.__path, '.lock')

        k = str(k)
        store_path = os.path.join(self.__path, k)
        return SimpleFileStore(store_path, binary=self.__binary, lock_path=lock_path)


    def ls(self):
        r = []
        for v in os.listdir(self.__path):
            if re.match(re_processedname, v):
                r.append(v)
        return r


    def close(self):
        pass



# local imports
from .base import StoreFactory


class NoopStore:

    def put(self, k, contents=None):
        pass


    def remove(self, k):
        pass

    def get(self, k):
        pass


    def list(self):
        return []


    def path(self):
        return None


    def replace(self, k, contents):
        pass

    def modified(self, k):
        pass


    def register_modify(self, k):
        pass


class NoopStoreFactory(StoreFactory):

    def add(self, k):
        return NoopStore()


    def ls(self):
        return []
